#! /usr/bin/python

# ---------------------- .spotdata files ------------------------
# spotdata files are csv files, each with the same number of fields
# ------------------------------------------------------------------

# ---------------- Instruction lines --------------
# 4 types:
# 1. blank
# 2. comment (leftmost non-space char is #)
# 3. label (rightmost non-space char is :)
# 4. instruction (command [arg1 [,arg2 [,arg3]]])
# ----------------------------- 

# ParseCSV(filename)
# returns dictionary with keys:
# rows (list of lists)
# numfields (number of fields)

from SpotInsline import *

# ------------------- lstins[] (list of instructions) -------------
lstins = []

# ParseInstructions(lst)
# returns dictionary with keys:
# isvalid (boolean)
# errormessage (if isvalid == False, then error message)
# lines[] (list of insline instances)


# -------------------------- ParseInstruction ------------------------
# parses instruction lines:  Instruction [arg1 [,arg2 ...]]
# line is presumed to be stripped, and not blank
# returns list and list-of-booleans (isquote)

def ParseInstruction(sline):
    where = sline.find(' ')
    if where < 0:
        # instruction only
        lst = [sline]
        isquote = [False]
        return lst, isquote
    
    lst = [sline[:where]]
    isquote = [False]
    
    args = sline[where+1:].strip()
    # if there are no quotes, it's easy
    if args.find('"') < 0:
        l = args.split(',')
        for i in range(len(l)):
            l[i] = l[i].strip()
        lst = lst + l
        isquote = len(lst) * [False]
        return lst, isquote
        
    #if, on the other hand, there are quotes, we have to go through it more carefully
    # if there are an odd number of quotes, add one at the end
    if args.count('"') % 2 ==1: args += '"'
    while args != '':
        #print 'args: '+args
        if args[0] == '"':
            where = args[1:].find('"')
            if where < 0:
                lst.append(args[1:])
                isquote = isquote + [True]
                break
            lst.append(args[1:where+1])
            isquote = isquote + [True]
            args = args[where+2:].lstrip()
            if args != '':
                if args[0] == ',':
                    args = args[1:].lstrip()
                else:
                    where = args[1:].find(',')
                    if where < 0:
                        break
                    args = args[where+1:].lstrip()
        else:
            where = args.find(',')
            if where < 0:
                lst.append(args)
                isquote = isquote + [False]
                break
            lst.append(args[:where])
            isquote = isquote + [False]
            args = args[where+1:].lstrip()
    return lst, isquote

# ---------------------------- ParseCommand ---------------------
# gets list of words on command line, isquote list and partially
# filled in insline, checks validity, and fills in insline
# returns errormessage if bad
def ParseCommand(lstwords, isquote, ins):
    errormessage = ''
    if lstwords[0] not in commandnames:
        return 'Unknown command: ' + lstwords[0]
    ins.command = lstwords[0]
    arglist = commands[lstwords[0]]
    if len(lstwords) != len(arglist[0]) + 1:
        if len(lstwords) < len(arglist[0]) + 1:
            errormessage = 'Too few arguments'
        else:
            errormessage = 'Too many arguments'
    else:
        for arglst1 in arglist:
            found = True
            for i in range(len(arglst1)):
                if arglst1[i] == 'string' and not isquote[i+1]: found = False; break
                if arglst1[i] == 'label' and not islabel(lstwords[i+1]): found = False; break
                if arglst1[i] == 'number' and not isnumber(lstwords[i+1]): found = False; break
                if arglst1[i] == 'card' and not iscard(lstwords[i+1]): found = False; break
                if arglst1[i] == 'special' and not isspecial(lstwords[i+1]): found = False; break
            if found:
                ins.args = {}
                ins.args['numargs'] = len(arglst1)
                for i in range(len(arglst1)):
                    ins.args['arg'+str(i)+'_type'] = arglst1[i]
                    s = 'arg' + str(i) + '_value'
                    if arglst1[i] == 'string' or arglst1[i] == 'label': ins.args[s] = lstwords[i+1]
                    if arglst1[i] == 'card': ins.args[s] = int(lstwords[i+1][1:])
                    if arglst1[i] == 'number': ins.args[s] = int(lstwords[i+1])
                    if arglst1[i] == 'special': ins.args[s] = lstwords[i+1]
                break
        if not found:
            errormessage = 'Incorrect type of arguments for this instruction'
    return errormessage

# ----------------------------- ParseOneLine ---------------------
# parse one instruction line, returns errormessage and insline
def ParseOneLine(sline):
    ins = insline()
    ins.sline = sline
    errormessage = ''
    sline = RemoveComments(sline)
    if sline == '':
        ins.type = 'blank'
    elif sline[0] == '#':
        ins.type = 'comment'
    elif sline[-1] == ':':
        if len(sline) < 2:
            errormessage = 'Label expected on this line'
        ins.type = 'label'
        ins.label = sline[:-1].lower()
    else:
        ins.type = 'command'
        # separate into words, keeping track of arguments in quotes
        lstwords,isquote = ParseInstruction(sline)
        for i in range(len(lstwords)):
            if not isquote[i]: lstwords[i] = lstwords[i].lower()
        # heavy lifting: check all command types and fill in insline
        errormessage = ParseCommand(lstwords, isquote, ins)
    return errormessage, ins

# format error message
def ErrorFormat(errormessage, sline):
    s = errormessage
    if sline != '': s = s+"<br>\nError in line:<br>\n"+sline
    return s

# CheckLabels(lst) checks labels for uniqueness and that
# all jump, gosub and jump commands have valid targets
# returns errormessage on failure, else ''
def CheckLabels(lst):
    # create a list of all of the labels
    labels = []
    for ins in lst:
        if ins.type == 'label':
            lab = ins.label.lower()
            if lab in labels: return ErrorFormat('Duplicate label',ins.sline)
            labels.append(lab)
    # now check that jump, gosub and jump all have valid targets
    for ins in lst:
        if ins.type == 'command':
            if ins.command[:5] in ('jump','gosub', 'jump-'):
                if ins.args['arg0_value'] not in labels:
                    return ErrorFormat('No target label found',ins.sline)
            
    return ''

# remove comments from a line
def RemoveComments(s):
    sout = s.strip()
    isharp = -1
    while True:
        isharp = sout.find('#',isharp+1)
        if isharp < 0: break
        if sout[:isharp].count('"')% 2 == 0:
            sout = sout[:isharp].strip()
            break
    return sout

# --------------------------- ParseInstructions ------------------
# given the list of instruction lines, parse them
# return errormessage
# create list of insline
def ParseInstructions(lst):
    global lstins
    ncommands = 0
    lstins = []
    for i in range(len(lst)):
        errormessage, ins = ParseOneLine(lst[i])
        if errormessage != '': return ErrorFormat(errormessage,lst[i])
        lstins.append(ins)
        if ins.type == 'command': ncommands += 1
    # any commands?
    if ncommands == 0: return ErrorFormat('No commands found.','')
    # check that all jump, gosub and jump commands have targets
    errormessage = CheckLabels(lstins)
    if errormessage != '': return errormessage
    #print 'len(lstins): '+str(len(lstins))
    return ''
    

